"use strict";

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

Object.defineProperty(exports, "__esModule", { value: true });
/**
 * convert key path string to key array
 * @private
 * @param {String} path
 * @returns {Array}
 * @example
 * pathToKeys("selector.to[0][11].value"); // ["selector", "to", 0, 11, "value"]
 */
exports.pathToKeys = function (path) {
    return path.replace(/([^\.])\[/g, function (m, match) {
        return match + ".[";
    }).split(".").map(function (key) {
        var match = key.match(/^\[(\d+)\]$/);
        if (match) return +match[1];
        return key;
    });
};
/**
 *
 * @private
 * @param {*} obj
 * @param {String} key
 * @param {*} value
 * @example
 * addPrototype(Object, "$size", function() {return this;});
 */
exports.addPrototype = function (obj, key, method) {
    var prototype = obj.prototype;
    if (prototype.hasOwnProperty(key)) return;
    Object.defineProperty(prototype, key, {
        value: function value() {
            for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
                args[_key] = arguments[_key];
            }

            return method.apply(0, [this].concat(args));
        },

        writable: true
    });
};
/**
 *
 * @private
 * @param {Array} arr
 * @param {String} [path]
 * @param {Function} fn
 * @returns {*}
 * @example
 * filter([1, 2, 3], undefined, (value) => value > 1);
 */
exports.filter = function (arr, path, fn) {
    if (path) {
        var keys = exports.pathToKeys(path);
        var reducer = function reducer(item) {
            return keys.reduce(function (prev, curr) {
                return prev[curr];
            }, item);
        };
        return arr.filter(function (item, index, items) {
            return fn(reducer(item), index, items);
        });
    }
    return arr.filter(fn);
};
/**
 *
 * @private
 * @param {*} value
 * @returns {*}
 * @example
 * deepClone([1, 2, 3]);
 */
exports.deepClone = function (value) {
    if ((typeof value === "undefined" ? "undefined" : _typeof(value)) !== "object") return value;
    if (value instanceof Date) return new Date(value.getTime());
    var clone = Object.assign({}, value);
    Object.keys(clone).forEach(function (key) {
        return clone[key] = _typeof(value[key]) === "object" ? exports.deepClone(value[key]) : value[key];
    });
    if (Array.isArray(value)) {
        clone.length = value.length;
        return Array.from(clone);
    }
    return clone;
};