"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * convert key path string to key array
 * @private
 * @param {String} path
 * @returns {Array}
 * @example
 * pathToKeys("selector.to[0][11].value"); // ["selector", "to", 0, 11, "value"]
 */
exports.pathToKeys = (path) => path
    .replace(/([^\.])\[/g, (m, match) => `${match}.[`)
    .split(".")
    .map(key => {
    const match = key.match(/^\[(\d+)\]$/);
    if (match)
        return +match[1];
    return key;
});
/**
 *
 * @private
 * @param {*} obj
 * @param {String} key
 * @param {*} value
 * @example
 * addPrototype(Object, "$size", function() {return this;});
 */
exports.addPrototype = (obj, key, method) => {
    const prototype = obj.prototype;
    if (prototype.hasOwnProperty(key))
        return;
    Object.defineProperty(prototype, key, {
        value(...args) {
            return method.apply(0, [this].concat(args));
        },
        writable: true,
    });
};
/**
 *
 * @private
 * @param {Array} arr
 * @param {String} [path]
 * @param {Function} fn
 * @returns {*}
 * @example
 * filter([1, 2, 3], undefined, (value) => value > 1);
 */
exports.filter = (arr, path, fn) => {
    if (path) {
        const keys = exports.pathToKeys(path);
        const reducer = (item) => keys.reduce((prev, curr) => prev[curr], item);
        return arr.filter((item, index, items) => fn(reducer(item), index, items));
    }
    return arr.filter(fn);
};
/**
 *
 * @private
 * @param {*} value
 * @returns {*}
 * @example
 * deepClone([1, 2, 3]);
 */
exports.deepClone = (value) => {
    if (typeof value !== "object")
        return value;
    if (value instanceof Date)
        return new Date(value.getTime());
    const clone = Object.assign({}, value);
    Object.keys(clone).forEach(key => (clone[key] =
        typeof value[key] === "object"
            ? exports.deepClone(value[key])
            : value[key]));
    if (Array.isArray(value)) {
        clone.length = value.length;
        return Array.from(clone);
    }
    return clone;
};
